atftp: Add files needed by the recipe

This commit is contained in:
Holger Hans Peter Freyther 2015-02-09 07:54:04 +01:00
parent b66bf237e9
commit cfc0c70237
5 changed files with 456 additions and 0 deletions

View File

@ -0,0 +1,94 @@
atftp exhibits the well known "Sorcerer's Apprentice Syndrome"(SAS) problem.
According to RFC 1350, the fix to SAS is quite simple: further copies of the
acknowledgment for a particular data block would be ignored.
Patch originally from OpenSUSE:
https://build.opensuse.org/package/view_file?file=atftp-0.7-sorcerers_apprentice.patch&package=atftp.539&project=openSUSE%3A12.1%3AUpdate&rev=84569792975e00573d7df597d2a6e895
Upstream-Status: Pending
Signed-off-by: Roy.Li <rongqing.li@windriver.com>
Index: atftp-0.7/tftp_file.c
===================================================================
--- atftp-0.7.orig/tftp_file.c 2011-11-22 15:12:53.792744083 +0100
+++ atftp-0.7/tftp_file.c 2011-11-22 15:13:51.706421893 +0100
@@ -605,6 +605,7 @@
int timeout_state = state; /* what state should we go on when timeout */
int result;
long block_number = 0;
+ long last_requested_block = -1;
long last_block = -1;
int data_size; /* size of data received */
int sockfd = data->sockfd; /* just to simplify calls */
@@ -765,6 +766,17 @@
connected = 1;
}
block_number = ntohs(tftphdr->th_block);
+
+ if (last_requested_block >= block_number)
+ {
+ if (data->trace)
+ fprintf(stderr, "received duplicated ACK <block: %ld >= %ld>\n",
+ last_requested_block, block_number);
+ break;
+ }
+ else
+ last_requested_block = block_number;
+
if (data->trace)
fprintf(stderr, "received ACK <block: %ld>\n",
block_number);
Index: atftp-0.7/tftpd_file.c
===================================================================
--- atftp-0.7.orig/tftpd_file.c 2011-11-22 15:12:53.793744112 +0100
+++ atftp-0.7/tftpd_file.c 2011-11-22 15:15:04.617534260 +0100
@@ -403,6 +403,7 @@
int timeout_state = state;
int result;
long block_number = 0;
+ long last_requested_block = -1;
long last_block = -1;
int block_loops = 0;
int data_size;
@@ -859,6 +860,32 @@
{
logger(LOG_DEBUG, "received ACK <block: %d>", block_number);
}
+
+ /* check whether the block request isn't already fulfilled */
+
+ /* multicast, block numbers could contain gaps */
+ if (multicast) {
+ if (last_requested_block >= block_number)
+ {
+ if (data->trace)
+ logger(LOG_DEBUG, "received duplicated ACK <block: %d >= %d>", last_requested_block, block_number);
+ break;
+ }
+ else
+ last_requested_block = block_number;
+ /* unicast, blocks should be requested one after another */
+ } else {
+ if (last_requested_block + 1 != block_number && last_requested_block != -1)
+ {
+ if (data->trace)
+ logger(LOG_DEBUG, "received out of order ACK <block: %d != %d>", last_requested_block + 1, block_number);
+ break;
+ }
+ else
+ last_requested_block = block_number;
+ }
+
+
if (ntohs(tftphdr->th_block) == 65535)
{
block_loops++;
@@ -958,6 +985,8 @@
/* nedd to send an oack to that client */
state = S_SEND_OACK;
fseek(fp, 0, SEEK_SET);
+ /* reset the last block received counter */
+ last_requested_block = -1;
}
else
{

View File

@ -0,0 +1,163 @@
Fate #303031: Circumvent TFTP size restrictions in atftpd
The size of a single image file that can be transferred with TFTP is limited to
2^(2*8) *BLOCKSIZE (as per RFC 1350 there are only two bytes for the block
counter). This is problematic for one of our customers who needs to transfer
100+ MB Windows images using a TFTP client (NT bootloader) which has a
hardwared BLOCKSIZE setting of 1432).
block rollover
http://www.compuphase.com/tftp.htm
Patch originally from OpenSUSE:
https://build.opensuse.org/package/show?package=atftp&project=openSUSE%3A12.2
Upstream-Status: Pending
Index: git/tftp_def.h
===================================================================
--- git.orig/tftp_def.h 2012-11-19 16:28:50.221027144 -0800
+++ git/tftp_def.h 2012-11-20 17:40:54.391206979 -0800
@@ -32,6 +32,7 @@
#define TIMEOUT 5 /* Client timeout */
#define S_TIMEOUT 5 /* Server timout. */
#define NB_OF_RETRY 5
+#define MAXBLOCKS 1000000 /* maximum number of blocks in a download */
/* definition to use tftp_options structure */
#define OPT_FILENAME 0
Index: git/tftp_file.c
===================================================================
--- git.orig/tftp_file.c 2012-11-19 16:28:50.221027144 -0800
+++ git/tftp_file.c 2012-11-19 16:28:51.201027167 -0800
@@ -622,8 +622,8 @@
int state = S_SEND_REQ; /* current state in the state machine */
int timeout_state = state; /* what state should we go on when timeout */
int result;
- int block_number = 0;
- int last_block = -1;
+ long block_number = 0;
+ long last_block = -1;
int data_size; /* size of data received */
int sockfd = data->sockfd; /* just to simplify calls */
struct sockaddr_storage sa; /* a copy of data.sa_peer */
@@ -637,8 +637,8 @@
int convert = 0; /* if true, do netascii convertion */
char string[MAXLEN];
- int prev_block_number = 0; /* needed to support netascii convertion */
- int prev_file_pos = 0;
+ long prev_block_number = 0; /* needed to support netascii convertion */
+ long prev_file_pos = 0;
int temp = 0;
data->file_size = 0;
@@ -745,7 +745,7 @@
data_size, data->data_buffer);
data->file_size += data_size;
if (data->trace)
- fprintf(stderr, "sent DATA <block: %d, size: %d>\n",
+ fprintf(stderr, "sent DATA <block: %ld, size: %d>\n",
block_number + 1, data_size - 4);
state = S_WAIT_PACKET;
break;
@@ -785,7 +785,7 @@
}
block_number = ntohs(tftphdr->th_block);
if (data->trace)
- fprintf(stderr, "received ACK <block: %d>\n",
+ fprintf(stderr, "received ACK <block: %ld>\n",
block_number);
if ((last_block != -1) && (block_number > last_block))
{
Index: git/tftp_io.c
===================================================================
--- git.orig/tftp_io.c 2012-11-19 16:28:50.221027144 -0800
+++ git/tftp_io.c 2012-11-19 16:28:51.201027167 -0800
@@ -350,8 +350,8 @@
/*
* Read from file and do netascii conversion if needed
*/
-int tftp_file_read(FILE *fp, char *data_buffer, int data_buffer_size, int block_number,
- int convert, int *prev_block_number, int *prev_file_pos, int *temp)
+int tftp_file_read(FILE *fp, char *data_buffer, int data_buffer_size, long block_number,
+ int convert, long *prev_block_number, long *prev_file_pos, int *temp)
{
int i;
int c;
Index: git/tftp_io.h
===================================================================
--- git.orig/tftp_io.h 2012-11-19 16:28:50.221027144 -0800
+++ git/tftp_io.h 2012-11-19 16:28:51.201027167 -0800
@@ -52,8 +52,8 @@
int tftp_get_packet(int sock1, int sock2, int *sock, struct sockaddr_storage *sa,
struct sockaddr_storage *from, struct sockaddr_storage *to,
int timeout, int *size, char *data);
-int tftp_file_read(FILE *fp, char *buffer, int buffer_size, int block_number, int convert,
- int *prev_block_number, int *prev_file_pos, int *temp);
+int tftp_file_read(FILE *fp, char *buffer, int buffer_size, long block_number, int convert,
+ long *prev_block_number, long *prev_file_pos, int *temp);
int tftp_file_write(FILE *fp, char *data_buffer, int data_buffer_size, int block_number,
int data_size, int convert, int *prev_block_number, int *temp);
#endif
Index: git/tftpd_list.c
===================================================================
--- git.orig/tftpd_list.c 2012-10-24 21:48:47.000000000 -0700
+++ git/tftpd_list.c 2012-10-24 21:52:04.266205076 -0700
@@ -49,11 +49,11 @@
*/
int tftpd_list_add(struct thread_data *new)
{
- struct thread_data *current = thread_data;
+ struct thread_data *current;
int ret;
pthread_mutex_lock(&thread_list_mutex);
-
+ current = thread_data;
number_of_thread++;
ret = number_of_thread;
@@ -81,11 +81,13 @@
*/
int tftpd_list_remove(struct thread_data *old)
{
- struct thread_data *current = thread_data;
+ struct thread_data *current;
int ret;
pthread_mutex_lock(&thread_list_mutex);
+ current = thread_data;
+
number_of_thread--;
ret = number_of_thread;
@@ -137,23 +139,26 @@
struct thread_data *data,
struct client_info *client)
{
- struct thread_data *current = thread_data; /* head of the list */
- struct tftp_opt *tftp_options = data->tftp_options;
+ struct thread_data *current; /* head of the list */
+ struct tftp_opt *tftp_options;
struct client_info *tmp;
char options[MAXLEN];
char string[MAXLEN];
char *index;
int len;
+ /* lock the whole list before walking it */
+ pthread_mutex_lock(&thread_list_mutex);
+
*thread = NULL;
+ current = thread_data;
+ tftp_options = data->tftp_options;
+
opt_request_to_string(tftp_options, options, MAXLEN);
index = strstr(options, "multicast");
len = (int)index - (int)options;
- /* lock the whole list before walking it */
- pthread_mutex_lock(&thread_list_mutex);
-
while (current)
{
if (current != data)
@@ -214,9 +219,10 @@
void tftpd_clientlist_remove(struct thread_data *thread,
struct client_info *client)
{
- struct client_info *tmp = thread->client_info;
+ struct client_info *tmp;
pthread_mutex_lock(&thread->client_mutex);
+ tmp = thread->client_info;
while ((tmp->next != client) && (tmp->next != NULL))
tmp = tmp->next;
if (tmp->next == NULL)
@@ -231,9 +237,11 @@
void tftpd_clientlist_free(struct thread_data *thread)
{
struct client_info *tmp;
- struct client_info *head = thread->client_info;
+ struct client_info *head;
pthread_mutex_lock(&thread->client_mutex);
+ head = thread->client_info;
+
while (head)
{
tmp = head;
@@ -250,9 +258,10 @@
struct client_info *client,
struct sockaddr_storage *sock)
{
- struct client_info *head = thread->client_info;
+ struct client_info *head;
pthread_mutex_lock(&thread->client_mutex);
+ head = thread->client_info;
if (client)
{
@@ -334,10 +343,10 @@
void tftpd_list_kill_threads(void)
{
- struct thread_data *current = thread_data; /* head of list */
+ struct thread_data *current; /* head of list */
pthread_mutex_lock(&thread_list_mutex);
-
+ current = thread_data;
while (current != NULL)
{
Index: git/tftpd_mcast.c
===================================================================
--- git.orig/tftpd_mcast.c 2012-10-24 21:48:47.000000000 -0700
+++ git/tftpd_mcast.c 2012-10-24 21:49:11.570201582 -0700
@@ -51,9 +51,11 @@
*/
int tftpd_mcast_get_tid(char **addr, short *port)
{
- struct tid *current = tid_list;
+ struct tid *current;
pthread_mutex_lock(&mcast_tid_list);
+ current = tid_list;
+
/* walk the list for a free tid */
while (current != NULL)
{
@@ -74,9 +76,11 @@
int tftpd_mcast_free_tid(char *addr, short port)
{
- struct tid *current = tid_list;
+ struct tid *current;
pthread_mutex_lock(&mcast_tid_list);
+ current = tid_list;
+
while (current != NULL)
{
if ((current->used == 1) && (current->port == port) &&

View File

@ -0,0 +1,37 @@
#! /bin/sh
#
# This is an init script for openembedded
# Copy it to /etc/init.d/atftpd and type
# > update-rc.d atftpd defaults 60
#
test -f /usr/sbin/atftpd || exit 0
test -d /srv/tftp || mkdir -p /srv/tftp
case "$1" in
start)
echo -n "Starting tftp daemon: atftpd"
start-stop-daemon --start --quiet --exec /usr/sbin/atftpd -- --daemon --port 69
echo "."
;;
stop)
echo -n "Stopping tftp daemon: atftpd"
start-stop-daemon --stop --quiet --exec /usr/sbin/atftpd
echo "."
;;
reload|force-reload)
start-stop-daemon --stop --quiet --signal 1 --exec /usr/sbin/atftpd
;;
restart)
$0 stop
sleep 1
$0 start
;;
*)
echo "Usage: /etc/init.d/atftpd {start|stop|reload|restart|force-reload}"
exit 1
esac
exit 0

View File

@ -0,0 +1,10 @@
[Unit]
Description=Advanced TFTP Server
After=syslog.target network.target
[Service]
Type=forking
ExecStart=/usr/sbin/atftpd --daemon --port 69
[Install]
WantedBy=multi-user.target
Index: git/tftpd_file.c
===================================================================
--- git.orig/tftpd_file.c 2012-11-19 16:28:50.225027144 -0800
+++ git/tftpd_file.c 2012-11-19 16:28:51.201027167 -0800
@@ -407,8 +407,9 @@
int state = S_BEGIN;
int timeout_state = state;
int result;
- int block_number = 0;
- int last_block = -1;
+ long block_number = 0;
+ long last_block = -1;
+ int block_loops = 0;
int data_size;
struct sockaddr_storage *sa = &data->client_info->client;
struct sockaddr_storage from;
@@ -431,8 +432,8 @@
struct client_info *client_old = NULL;
struct tftp_opt options[OPT_NUMBER];
- int prev_block_number = 0; /* needed to support netascii convertion */
- int prev_file_pos = 0;
+ long prev_block_number = 0; /* needed to support netascii convertion */
+ long prev_file_pos = 0;
int temp = 0;
/* look for mode option */
@@ -565,11 +566,12 @@
logger(LOG_INFO, "blksize option -> %d", result);
}
- /* Verify that the file can be sent in 2^16 block of BLKSIZE octets */
- if ((file_stat.st_size / (data->data_buffer_size - 4)) > 65535)
+ /* Verify that the file can be sent in MAXBLOCKS blocks of BLKSIZE octets */
+ if ((file_stat.st_size / (data->data_buffer_size - 4)) > MAXBLOCKS)
{
tftp_send_error(sockfd, sa, EUNDEF, data->data_buffer, data->data_buffer_size);
- logger(LOG_NOTICE, "Requested file to big, increase BLKSIZE");
+ logger(LOG_NOTICE, "Requested file too big, increase BLKSIZE");
+ logger(LOG_NOTICE, "Only %d blocks of %d bytes can be served.", MAXBLOCKS, data->data_buffer_size);
if (data->trace)
logger(LOG_DEBUG, "sent ERROR <code: %d, msg: %s>", EUNDEF,
tftp_errmsg[EUNDEF]);
@@ -880,10 +882,15 @@
}
/* The ACK is from the current client */
number_of_timeout = 0;
- block_number = ntohs(tftphdr->th_block);
+ block_number = (block_loops * 65536) + ntohs(tftphdr->th_block);
if (data->trace)
- logger(LOG_DEBUG, "received ACK <block: %d>",
- block_number);
+ {
+ logger(LOG_DEBUG, "received ACK <block: %d>", block_number);
+ }
+ if (ntohs(tftphdr->th_block) == 65535)
+ {
+ block_loops++;
+ };
if ((last_block != -1) && (block_number > last_block))
{
state = S_END;

View File

@ -0,0 +1,152 @@
Avoid assigning thread data outside of mutex lock
Patch originally from OpenSUSE:
https://build.opensuse.org/package/show?package=atftp&project=openSUSE%3A12.2
Upstream-Status: Pending